(function() {
    'use strict';

    angular.module('weChatApp')
        .factory('Endpoints', ["$q", "Queries", Endpoints])
    ;

    function Endpoints($q, queries) {
        var endpoints = this;
        var utils = new window.VeevaUtilities();

        var normalizeIdForOnline = function(responseObject) {
            if(angular.isObject(responseObject) && angular.isDefined(responseObject.ID)) {
                responseObject.Id = responseObject.ID;
            }
        };

        var normalizeWhereClauseBooleanValue = function(value) {
            if(utils.isWin8() && !utils.isOnline()) {
                return (typeof value === 'boolean' && value) ? '\'true\'' : '\'false\'';
            }
            return typeof value === 'boolean' && value;
        };

        var normalizeWhereClauseDateValue = function(value) {
            if(utils.isOnline()) {
                return value;
            } else {
                return "'" + value + "'";
            }
        };

        var normalizeWhereClauseEmptyValue = function() {
            if(utils.isOnline()) {
                return null;
            } else {
                return "''";
            }
        };

        function genericQueryErrorHandler(deferred) {
            return function(error) {
                deferred.reject(error);
            };
        }

        endpoints.fetchCurrentUserId = function() {
            var deferred = $q.defer();

            ds.getDataForCurrentObject("User", "Id").then(function(resp) {
                if(resp.success) {
                    if(utils.isOnline()) {
                        normalizeIdForOnline(resp.User);
                    }
                    deferred.resolve(resp.User.Id);
                } else {
                    (genericQueryErrorHandler(deferred))(resp.message);
                }
            }, genericQueryErrorHandler(deferred));

            return deferred.promise;
        };

        endpoints.fetchUserById = function(userId) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.userById);
            queryConfig.where = queryConfig.where.replace(/_id/, '\'' + userId + '\'');
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchUserByIds = function(userIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.userByIds);
            queryConfig.where += ds.getInStatement(userIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchTerritoryByUser = function(userId) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.territoryByUserId);
            queryConfig.where = queryConfig.where.replace(/_userId/, '\'' + userId + '\'');
            queryConfig.where = queryConfig.where.replace(/_true/, normalizeWhereClauseBooleanValue(true));
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchTerritory2ByUser = function(userId) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.territory2ByUserId);
            queryConfig.where = queryConfig.where.replace(/_userId/, '\'' + userId + '\'');
            queryConfig.where = queryConfig.where.replace(/_true/, normalizeWhereClauseBooleanValue(true));
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAllTerritories = function() {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.territory);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAllTerritories2 = function() {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.territory2);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAvailableObjects = function() {
            var deferred = $q.defer();
            ds.getAvailableObjects().then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchObjectMetadata = function(objectName) {
            var deferred = $q.defer();
            ds.getObjectMetadata({object: objectName}).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchVeevaMessages = function(msgList, locale) {
            var deferred = $q.defer();
            ds.getVeevaMessagesWithDefault(msgList, locale).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchFieldLabels = function(objectName, fieldNames) {
            var deferred = $q.defer();
            ds.getFieldLabels({object: objectName, fields: fieldNames}).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchObjectLabels = function(objectNames) {
            var deferred = $q.defer();
            ds.getObjectLabels(objectNames).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchPicklistLabels = function(objectName, fieldName) {
            var deferred = $q.defer();
            ds.getPicklistValueLabels(objectName, fieldName).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchRecordTypeLabels = function(objectName) {
            var deferred = $q.defer();
            ds.getRecordTypeLabels(objectName).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchSentMessagesByAccountIds = function(accountIds, startDatetime, endDatetime) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.sentMessagesByAccount);
            queryConfig.where = queryConfig.where.replace(/_datetime1/, normalizeWhereClauseDateValue(startDatetime));
            queryConfig.where = queryConfig.where.replace(/_datetime2/, normalizeWhereClauseDateValue(endDatetime));
            queryConfig.where += ds.getInStatement(accountIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAllSentMessagesByAccountIds = function(accountIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.sentMessagesByAccount);
            queryConfig.where = "Account_vod__c IN " + ds.getInStatement(accountIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchMultichannelConsentByAccountIds = function(accountIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.multichannelConsentByAccountIds);
            queryConfig.where += ds.getInStatement(accountIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAccountByIds = function(accountIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.accountByIds);
            queryConfig.where += ds.getInStatement(accountIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAccountByTerritoryIds = function(territoryIds) {
            var deferred = $q.defer();

            // To get Account by TerritoryIds, we need to first get the territory related id from Group
            endpoints.fetchTerritoryGroupByRelatedId(territoryIds).then(function(groupResp) {
                // Then we can query account ids from AccountShare by the territory related id
                endpoints.fetchAccountShareByGroupIds(_.map(groupResp.Group, "Id")).then(function(accountShareResp) {
                    // Finally, query account object once we get the account ids from AccountShare
                    endpoints.fetchAccountByIds(_.map(accountShareResp.AccountShare, "AccountId")).then(function(accountResp) {
                        // Add territory data for each account
                        var accountWithTerritory = accountResp.Account.map(function(acct) {
                            // for each account, trace back to its accountShare record
                            var accountShareRecordsForAccountX = accountShareResp.AccountShare.filter(function(as) {
                                return as.AccountId === acct.Id;
                            });
                            // get the group ids found inside the accountShare records
                            var groupIdsForAccountX = _.map(accountShareRecordsForAccountX, "UserOrGroupId");
                            // filter the group records to only include the given group ids from above
                            var groupRecordsForAccountX = groupResp.Group.filter(function(g) {
                                return _.includes(groupIdsForAccountX, g.Id)
                            });
                            // get the territory ids from group records
                            acct._territoryIds = _.map(groupRecordsForAccountX, "RelatedId");
                            return acct;
                        });
                        deferred.resolve(accountWithTerritory);
                    }, genericQueryErrorHandler(deferred));
                }, genericQueryErrorHandler(deferred));
            }, genericQueryErrorHandler(deferred));

            return deferred.promise;
        };

        endpoints.fetchAccountByTerritory2Ids = function(territoryIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.OT2A);

            queryConfig.where += ds.getInStatement(territoryIds);

            ds.queryRecord(queryConfig).then(function(resp){
                var accountIds = [];
                var accountTerritoryMap = {};

                if (resp && angular.isArray(resp.ObjectTerritory2Association)) {
                    accountIds = resp.ObjectTerritory2Association.map(function(element) {
                        accountTerritoryMap[element.ObjectId] = [];
                        accountTerritoryMap[element.ObjectId].push(element.Territory2Id);
                        return element.ObjectId;
                    });
                }

                endpoints.fetchAccountByIds(accountIds).then(function(accountResp) {
                    // Add territory data for each account
                    var accountWithTerritory = accountResp.Account.map(function(acct) {
                        acct['_territoryIds'] = accountTerritoryMap[acct.Id];
                        return acct;
                    });
                    deferred.resolve(accountWithTerritory);
                }, genericQueryErrorHandler(deferred));
            }, genericQueryErrorHandler(deferred));

            return deferred.promise;
        };

        endpoints.fetchTerritoryGroupByRelatedId = function(relatedIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.territoryGroupByRelatedIds);
            queryConfig.where += ds.getInStatement(relatedIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchAccountShareByGroupIds = function(groupIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.accountShareByGroupIds);
            queryConfig.where += ds.getInStatement(groupIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchMultichannelContentByIds = function(mcContentIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.multichannelContentByIds);
            queryConfig.where += ds.getInStatement(mcContentIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchContentTypeByIds = function(contentTypeIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.contentTypeByIds);
            queryConfig.where += ds.getInStatement(contentTypeIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchWeChatRecordType = function() {
            var deferred = $q.defer();
            ds.queryRecord(queries.weChatRecordType).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchTargetTsfRecordByExternalIds = function(externalIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.tsfByExtId);
            queryConfig.where = queryConfig.where.replace(/_true/, normalizeWhereClauseBooleanValue(true));
            queryConfig.where += ds.getInStatement(externalIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchUsersByTerritoryIds = function(territoryIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.userByTerritoryIds);
            queryConfig.where += ds.getInStatement(territoryIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        endpoints.fetchUsersByTerritory2Ids = function(territoryIds) {
            var deferred = $q.defer();
            var queryConfig = angular.copy(queries.userByTerritory2Ids);
            queryConfig.where += ds.getInStatement(territoryIds);
            ds.queryRecord(queryConfig).then(deferred.resolve, genericQueryErrorHandler(deferred));
            return deferred.promise;
        };

        return endpoints;
    }
}());
